
"""
Main Window of the TuiView application. Contains
the ViewerWidget, menus, toolbars and status bars.
"""
# This file is part of 'TuiView' - a simple Raster viewer
# Copyright (C) 2012  Sam Gillingham
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 2
# of the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.

import os
import sys
import traceback
from PyQt5.QtWidgets import QMainWindow, QAction, QFileDialog, QDialog
from PyQt5.QtWidgets import QMessageBox, QProgressBar, QToolButton
from PyQt5.QtWidgets import QMenu
from PyQt5.QtGui import QIcon, QColor
from PyQt5.QtCore import QSettings, QSize, QPoint, pyqtSignal, Qt
from PyQt5.QtCore import QCoreApplication, QEventLoop

# different resource file needed on Python3
if sys.version_info[0] == 3:
    from . import viewerresources3
else:
    from . import viewerresources2
from . import archivereader
from . import viewerwidget
from . import viewererrors
from . import querywindow
from .viewerstrings import MESSAGE_TITLE

# set to True to see traceback when file open fails
SHOW_TRACEBACK = (os.getenv('TUIVIEW_SHOW_TRACEBACK','0') == '1')

DEFAULT_XSIZE = 400
DEFAULT_YSIZE = 400
DEFAULT_XPOS = 200
DEFAULT_YPOS = 200

MESSAGE_TIMEOUT = 2000
DEFAULT_DRIVER = os.getenv('TUIVIEW_DFLT_DRIVER')

# Populate this list the first time the
# file open dialog shown.
GDAL_FILTERS = None

# Set up a dictionary of filters not in GDAL (ENVI files with various extensions)
NON_GDAL_FILTERS = {'BIL' : 'ENVI BIL (*.bil)',
                    'BSQ' : 'ENVI BSQ (*.bsq)',
                    'DEM' : 'ENVI DEM (*.dem)',
                    'RAW' : 'ENVI RAW (*.raw)'}

def createFilter(driver):
    """
    Given a GDAL driver, creates the Qt QFileDialog
    compatible filter for the file type
    """
    from osgeo.gdal import DMD_LONGNAME, DMD_EXTENSION
    drivermeta = driver.GetMetadata()
    name = 'Image Files'
    if DMD_LONGNAME in drivermeta:
        name = drivermeta[DMD_LONGNAME]
        # get rid of any stuff in brackets - seems to
        # confuse Qt 4.x
        firstbracket = name.find('(')
        if firstbracket != -1:
            name = name[:firstbracket]
    qfilter = '*'
    if DMD_EXTENSION in drivermeta:
        qfilter = drivermeta[DMD_EXTENSION]
    return "%s (*.%s)" % (name, qfilter)

def populateFilters(defaultDriver=DEFAULT_DRIVER):
    """
    Create a list of file filters for QFileDialog for
    all the GDAL supported files.
    If a default driver is specified it goes first on the list
    """
    from osgeo import gdal
    global GDAL_FILTERS
    # only bother if it hasn't been populated already
    if GDAL_FILTERS is None:
        GDAL_FILTERS = []

        # if we have a default driver do it first
        if not defaultDriver is None:
            driver = gdal.GetDriverByName(defaultDriver)
            if driver is not None:
                qfilter = createFilter(driver)
                GDAL_FILTERS.append(qfilter)
            else:
                # If there is no GDAL driver try non-GDAL drivers dict
                try:
                    qfilter = NON_GDAL_FILTERS[defaultDriver]
                    GDAL_FILTERS.append(qfilter)
                except KeyError:
                    pass

        # add all files next
        GDAL_FILTERS.append("All files (*)")

        # just go thru them all and create filters
        for count in range(gdal.GetDriverCount()):
            driver = gdal.GetDriver(count)
            # we have already done the default driver
            # and it looks a bit silly if it is in there again
            if defaultDriver is None or driver.ShortName != defaultDriver:
                qfilter = createFilter(driver)
                GDAL_FILTERS.append(qfilter)

        # Now add non-GDAL filters
        for qfilter in NON_GDAL_FILTERS.values():
            GDAL_FILTERS.append(qfilter)

class ViewerWindow(QMainWindow):
    """
    Main window for viewer application. The ViewerWidget is 
    contained in the 'viewwidget' attribute.
    """
    # signals
    newWindowSig = pyqtSignal(name='newWindow')
    tileWindowsSig = pyqtSignal(int, int, name='tileWindows')
    newQueryWindowSig = pyqtSignal(querywindow.QueryDockWidget,
                            name='newQueryWindow')
    # Don't know how to specify file objects...
    writeViewersState = pyqtSignal(object, name='writeViewersState')
    readViewersState = pyqtSignal(object, name='readViewersState')

    def __init__(self):
        QMainWindow.__init__(self)
        self.setWindowTitle(MESSAGE_TITLE)
        self.viewwidget = viewerwidget.ViewerWidget(self)

        # connect to the signals emmitted by the LUT/RAT via the LayerManager
        # so we can update our progress bar
        self.viewwidget.layers.newProgressSig.connect(self.newProgress)
        self.viewwidget.layers.endProgressSig.connect(self.endProgress)
        self.viewwidget.layers.newPercentSig.connect(self.newPercent)
        # so we can update the window title
        self.viewwidget.layers.topLayerChanged.connect(self.updateWindowTitle)
        # general messages from the widget
        self.viewwidget.showStatusMessage.connect(self.showStatusMessage)
        # the signal that gets sent when active tool changed so we can update
        # gui if querywindow engages a tool
        self.viewwidget.activeToolChanged.connect(self.activeToolChanged)

        self.setCentralWidget(self.viewwidget)

        self.setupActions()
        self.setupMenus()
        self.setupToolbars()
        self.setupStatusBar()

        # our layer window so we can toggle it
        self.layerWindow = None

        self.restoreFromSettings()
        # set these values, just read from settings
        self.viewwidget.setMouseScrollWheelAction(self.mouseWheelZoom)
        self.viewwidget.setBackgroundColor(self.backgroundColor)
        if self.settingQueryOnlyDisplayed:
            self.queryOnlyDisplayedAct.setChecked(True)
        if self.settingArrangeLayersOpen:
            self.arrangeLayers()

        self.showStatusMessage("Ready")

        # number of query windows we have open.
        # if zero we need to start a new one when query
        # tool selected
        self.queryWindowCount = 0
        # same, but for profile window
        self.profileWindowCount = 0
        # same, but for vector query window
        self.vectorQueryWindowCount = 0

        # accept dropping files
        self.setAcceptDrops(True)

        # so if we are turning on a tool because another tool 
        # in another window has been turned on, we don't undo 
        # that tool being enabled. As oppossed to user unclicking
        # the tool
        self.suppressToolReset = False

    def resizeForWidgetSize(self, xsize, ysize):
        """
        Resizes this window so that the widget is the given size
        Takes into account the border etc
        """
        viewerSize = self.size()
        widgetSize = self.viewwidget.viewport().size()
        borderWidth = viewerSize.width() - widgetSize.width()
        borderHeight = viewerSize.height() - widgetSize.height()

        # resize it to desired size
        self.resize(xsize + borderWidth, ysize + borderHeight)
        
    def updateWindowTitle(self, layer):
        """
        called in response to the topLayerChanged(PyQt_PyObject) signal
        from the layers to say the top displayed layer has changed
        """
        if not layer is None:
            # just display the layer title
            self.setWindowTitle(layer.title)
        else:
            # nothing loaded
            self.setWindowTitle(MESSAGE_TITLE)

    def newProgress(self, string):
        """
        Called when we are about to start a new progress
        """
        self.statusBar().showMessage(string)
        self.progressWidget.setValue(0)
        self.progressWidget.setVisible(True)
        self.setCursor(Qt.WaitCursor)  # look like we are busy
        # process any events show gets shown while busy
        QCoreApplication.processEvents(QEventLoop.ExcludeUserInputEvents)

    def endProgress(self):
        """
        Called when a progress run has finished
        """
        self.statusBar().clearMessage()
        self.progressWidget.setVisible(False)
        self.setCursor(Qt.ArrowCursor)  # look like we are finished
        # process any events show gets shown while busy
        QCoreApplication.processEvents(QEventLoop.ExcludeUserInputEvents)

    def newPercent(self, percent):
        """
        New progress value
        """
        self.progressWidget.setValue(percent)
        # process any events show gets shown while busy
        QCoreApplication.processEvents(QEventLoop.ExcludeUserInputEvents)

    def showStatusMessage(self, message):
        """
        Helper method to show a message for a short period of time
        """
        self.statusBar().showMessage(message, MESSAGE_TIMEOUT)

    def activeToolChanged(self, obj):
        """
        Called when the active tool changed. If we didn't cause it
        then show our tools as disabled
        """
        if obj.senderid != id(self):
            self.suppressToolReset = True
            for tool in self.toolActions:
                tool.setChecked(False)
            self.suppressToolReset = False

    def restoreFromSettings(self):
        """
        Restore any settings from last time
        n.b. need to rationalize with preferences window
        """
        settings = QSettings()
        settings.beginGroup('ViewerWindow')

        defaultsize = QSize(DEFAULT_XSIZE, DEFAULT_YSIZE)
        self.resize(settings.value("size", defaultsize))

        defaultpos = QPoint(DEFAULT_XPOS, DEFAULT_YPOS)
        self.move(settings.value("pos", defaultpos))

        settings.endGroup()

        settings.beginGroup('ViewerMouse')
        value = settings.value("mousescroll", True, bool)
        self.mouseWheelZoom = value
        settings.endGroup()

        settings.beginGroup('ViewerBackground')
        value = settings.value("color", QColor(Qt.black), QColor)
        self.backgroundColor = value
        settings.endGroup()

        settings.beginGroup('StartupState')
        value = settings.value('QueryOnlyDisplayed', False, bool)
        self.settingQueryOnlyDisplayed = value

        value = settings.value('ArrangeLayersOpen', False, bool)
        self.settingArrangeLayersOpen = value
        settings.endGroup()

    def setupActions(self):
        """
        Creates all the actions for the Window
        """
        self.toolActions = []

        self.addRasterAct = QAction(self, triggered=self.addRaster)
        self.addRasterAct.setText("&Add Raster")
        self.addRasterAct.setStatusTip("Open a GDAL supported image")
        self.addRasterAct.setShortcut("CTRL+O")
        self.addRasterAct.setIcon(QIcon(":/viewer/images/addraster.png"))
        self.addRasterAct.setIconVisibleInMenu(True)

        self.addVectorFileAct = QAction(self, triggered=self.addVectorFile)
        self.addVectorFileAct.setText("Add Vector &File")
        self.addVectorFileAct.setStatusTip("Open an OGR supported vector file")
        self.addVectorFileAct.setShortcut("CTRL+V")
        self.addVectorFileAct.setIcon(QIcon(":/viewer/images/addvector.png"))
        self.addVectorFileAct.setIconVisibleInMenu(True)

        self.addVectorDirAct = QAction(self, triggered=self.addVectorDir)
        self.addVectorDirAct.setText("Add Vector &Directory")
        self.addVectorDirAct.setStatusTip("Open an OGR supported vector directory")
        self.addVectorDirAct.setIcon(QIcon(":/viewer/images/addvector.png"))
        self.addVectorDirAct.setIconVisibleInMenu(True)

        self.addVectorDBAct = QAction(self, triggered=self.addVectorDB)
        self.addVectorDBAct.setText("Add Vector Data&base")
        self.addVectorDBAct.setStatusTip(
                                "Open a layer from an OGR supported database")
        self.addVectorDBAct.setIcon(QIcon(":/viewer/images/addvector.png"))
        self.addVectorDBAct.setIconVisibleInMenu(True)

        self.vectorMenu = QMenu()
        self.vectorMenu.setTitle("Add Vector")
        self.vectorMenu.addAction(self.addVectorFileAct)
        self.vectorMenu.addAction(self.addVectorDirAct)
        self.vectorMenu.addAction(self.addVectorDBAct)

        self.removeLayerAct = QAction(self, triggered=self.removeLayer)
        self.removeLayerAct.setText("&Remove Layer")
        self.removeLayerAct.setStatusTip("Remove top layer")
        self.removeLayerAct.setShortcut("CTRL+R")
        self.removeLayerAct.setIcon(QIcon(":/viewer/images/removelayer.png"))
        self.removeLayerAct.setIconVisibleInMenu(True)

        self.newWindowAct = QAction(self, triggered=self.newWindow)
        self.newWindowAct.setText("&New Window")
        self.newWindowAct.setStatusTip("Create a new geo linked window")
        self.newWindowAct.setShortcut("CTRL+N")
        self.newWindowAct.setIcon(QIcon(":/viewer/images/newwindow.png"))
        self.newWindowAct.setIconVisibleInMenu(True)

        self.tileWindowsAct = QAction(self, triggered=self.tileWindows)
        self.tileWindowsAct.setText("&Tile Windows...")
        self.tileWindowsAct.setStatusTip("Tile all open windows")
        self.tileWindowsAct.setShortcut("CTRL+I")

        self.defaultStretchAct = QAction(self, triggered=self.defaultStretch)
        self.defaultStretchAct.setText("&Default Stretch...")
        self.defaultStretchAct.setStatusTip("Set default stretches")
        self.defaultStretchAct.setShortcut("CTRL+D")

        self.stretchAct = QAction(self, triggered=self.editStretch)
        self.stretchAct.setText("S&tretch")
        self.stretchAct.setStatusTip("Edit current stretch")
        self.stretchAct.setShortcut("CTRL+T")
        self.stretchAct.setEnabled(False) # until a file is opened

        self.panAct = QAction(self, toggled=self.pan)
        self.panAct.setText("&Pan")
        self.panAct.setStatusTip("Pan")
        self.panAct.setShortcut("CTRL+P")
        self.panAct.setCheckable(True)
        self.panAct.setIcon(QIcon(":/viewer/images/pan.png"))
        self.panAct.setIconVisibleInMenu(True)
        self.toolActions.append(self.panAct)

        self.zoomInAct = QAction(self, toggled=self.zoomIn)
        self.zoomInAct.setText("Zoom &In")
        self.zoomInAct.setStatusTip("Zoom In")
        self.zoomInAct.setShortcut("CTRL++")
        self.zoomInAct.setCheckable(True)
        self.zoomInAct.setIcon(QIcon(":/viewer/images/zoomin.png"))
        self.zoomInAct.setIconVisibleInMenu(True)
        self.toolActions.append(self.zoomInAct)

        self.zoomOutAct = QAction(self, toggled=self.zoomOut)
        self.zoomOutAct.setText("Zoom &Out")
        self.zoomOutAct.setStatusTip("Zoom Out")
        self.zoomOutAct.setShortcut("CTRL+-")
        self.zoomOutAct.setCheckable(True)
        self.zoomOutAct.setIcon(QIcon(":/viewer/images/zoomout.png"))
        self.zoomOutAct.setIconVisibleInMenu(True)
        self.toolActions.append(self.zoomOutAct)

        self.zoomNativeAct = QAction(self, triggered=self.zoomNative)
        self.zoomNativeAct.setText("Zoom to &Native")
        self.zoomNativeAct.setStatusTip("Zoom to Native Resolution")
        self.zoomNativeAct.setShortcut("CTRL+1")
        self.zoomNativeAct.setIcon(QIcon(":/viewer/images/zoomnative.png"))
        self.zoomNativeAct.setIconVisibleInMenu(True)

        self.zoomFullExtAct = QAction(self, triggered=self.zoomFullExtent)
        self.zoomFullExtAct.setText("Zoom to &Full Extent")
        self.zoomFullExtAct.setStatusTip("Zoom to Full Extent")
        self.zoomFullExtAct.setShortcut("CTRL+F")
        self.zoomFullExtAct.setIcon(QIcon(":/viewer/images/zoomfullextent.png"))
        self.zoomFullExtAct.setIconVisibleInMenu(True)

        self.followExtentAct = QAction(self, toggled=self.followExtent)
        self.followExtentAct.setText("Follow &Extent")
        self.followExtentAct.setStatusTip("Follow geolinked extent")
        self.followExtentAct.setShortcut("CTRL+E")
        self.followExtentAct.setCheckable(True)
        self.followExtentAct.setChecked(True) # by default to match viewerwidget
        self.followExtentAct.setIcon(QIcon(":/viewer/images/followextents.png"))
        self.followExtentAct.setIconVisibleInMenu(True)

        self.queryAct = QAction(self, toggled=self.query)
        self.queryAct.setText("&Query Tool")
        self.queryAct.setStatusTip("Start Query Tool")
        self.queryAct.setShortcut("CTRL+U")
        self.queryAct.setCheckable(True)
        self.queryAct.setIcon(QIcon(":/viewer/images/query.png"))
        self.queryAct.setIconVisibleInMenu(True)

        self.newQueryAct = QAction(self, triggered=self.newQueryWindow)
        self.newQueryAct.setText("New Query &Window")
        self.newQueryAct.setStatusTip("Open New Query Window")
        self.newQueryAct.setShortcut("CTRL+W")

        self.vectorQueryAct = QAction(self, toggled=self.vectorQuery)
        self.vectorQueryAct.setText("&Vector Query Tool")
        self.vectorQueryAct.setStatusTip("Start Vector Query Tool")
        self.vectorQueryAct.setShortcut("CTRL+C")
        self.vectorQueryAct.setCheckable(True)
        self.vectorQueryAct.setIcon(QIcon(":/viewer/images/queryvector.png"))
        self.vectorQueryAct.setIconVisibleInMenu(True)

        self.newVectorQueryAct = QAction(self, 
                                        triggered=self.newVectorQueryWindow)
        self.newVectorQueryAct.setText("New Vector Query &Window")
        self.newVectorQueryAct.setStatusTip("Open New Vector Query Window")

        self.queryOnlyDisplayedAct = QAction(self, 
                                        toggled=self.queryOnlyDisplayed)
        self.queryOnlyDisplayedAct.setText("&Query Only Displayed Layers")
        self.queryOnlyDisplayedAct.setCheckable(True)
        self.queryOnlyDisplayedAct.setShortcut("CTRL+B")
        self.queryOnlyDisplayedAct.setStatusTip(
            "Query Only Displayed Layers with Query Window")

        self.exitAct = QAction(self, triggered=self.close)
        self.exitAct.setText("&Close")
        self.exitAct.setStatusTip("Close this window")
        self.exitAct.setShortcut("CTRL+Q")

        self.preferencesAct = QAction(self, triggered=self.setPreferences)
        self.preferencesAct.setText("&Preferences")
        self.preferencesAct.setStatusTip("Edit Preferences")
        self.preferencesAct.setShortcut("CTRL+L")

        self.flickerAct = QAction(self, triggered=self.flicker)
        self.flickerAct.setText("&Flicker")
        self.flickerAct.setStatusTip("Flicker top 2 layers")
        self.flickerAct.setShortcut("CTRL+K")
        self.flickerAct.iconOn = QIcon(":/viewer/images/flickeron.png")
        self.flickerAct.iconOff = QIcon(":/viewer/images/flickeroff.png")
        self.flickerAct.setIcon(self.flickerAct.iconOn)
        self.flickerAct.setIconVisibleInMenu(True)

        self.layerAct = QAction(self, triggered=self.arrangeLayers)
        self.layerAct.setText("Arrange La&yers")
        self.layerAct.setStatusTip("Arrange Layers")
        self.layerAct.setShortcut("CTRL+Y")
        self.layerAct.setIcon(QIcon(":/viewer/images/layers.png"))
        self.layerAct.setIconVisibleInMenu(True)

        self.profileAct = QAction(self, toggled=self.profile)
        self.profileAct.setText("&Profile/Ruler")
        self.profileAct.setStatusTip("Start Profile/Ruler tool")
        self.profileAct.setShortcut("CTRL+A")
        self.profileAct.setCheckable(True)
        self.profileAct.setIcon(QIcon(":/viewer/images/profileruler.png"))
        self.profileAct.setIconVisibleInMenu(True)
        self.toolActions.append(self.profileAct)

        self.newProfileAct = QAction(self, triggered=self.newProfile)
        self.newProfileAct.setText("New P&rofile/Ruler Window")
        self.newProfileAct.setStatusTip("Open New Profile/Ruler Window")
        self.newProfileAct.setShortcut("CTRL+S")

        self.propertiesAct = QAction(self, triggered=self.properties)
        self.propertiesAct.setText("Properties")
        self.propertiesAct.setStatusTip("Show Properties of top layer")
        self.propertiesAct.setShortcut("CTRL+X")
        self.propertiesAct.setIcon(QIcon(":/viewer/images/properties.png"))
        self.propertiesAct.setIconVisibleInMenu(True)

        self.timeseriesForwardAct = QAction(self, 
                        triggered=self.viewwidget.timeseriesForward)
        self.timeseriesForwardAct.setShortcut(".")
        self.timeseriesForwardAct.setText("Timeseries Forward")
        self.timeseriesForwardAct.setStatusTip(
                                "Go forward through timeseries of images")

        self.timeseriesBackwardAct = QAction(self, 
                        triggered=self.viewwidget.timeseriesBackward)
        self.timeseriesBackwardAct.setShortcut(",")
        self.timeseriesBackwardAct.setText("Timeseries Backward")
        self.timeseriesBackwardAct.setStatusTip(
                                "Go backward through timeseries of images")

        self.saveCurrentViewAct = QAction(self, triggered=self.saveCurrentView)
        self.saveCurrentViewAct.setText("Save Current Display")
        self.saveCurrentViewAct.setStatusTip(
                    "Save the contents of the current display as an image file")

        self.saveCurrentViewersState = QAction(self, 
                            triggered=self.saveViewersState)
        self.saveCurrentViewersState.setText("Save State of All Viewers")
        self.saveCurrentViewersState.setStatusTip(
                "Save state of Viewers to a file so they can be restored")

        self.loadCurrentViewersState = QAction(self, 
                            triggered=self.loadViewersState)
        self.loadCurrentViewersState.setText("Load State of Viewers")
        self.loadCurrentViewersState.setStatusTip(
                "Restore state of viewers previously saved")

        self.aboutAct = QAction(self, triggered=self.about)
        self.aboutAct.setText("&About")
        self.aboutAct.setStatusTip("Show author and version information")


        # Actions just for keyboard shortcuts

        self.moveUpAct = QAction(self, triggered=self.moveUp)
        self.moveUpAct.setShortcut(Qt.CTRL + Qt.Key_Up)

        self.moveDownAct = QAction(self, triggered=self.moveDown)
        self.moveDownAct.setShortcut(Qt.CTRL + Qt.Key_Down)

        self.moveLeftAct = QAction(self, triggered=self.moveLeft)
        self.moveLeftAct.setShortcut(Qt.CTRL + Qt.Key_Left)

        self.moveRightAct = QAction(self, triggered=self.moveRight)
        self.moveRightAct.setShortcut(Qt.CTRL + Qt.Key_Right)

        self.addAction(self.moveUpAct)
        self.addAction(self.moveDownAct)
        self.addAction(self.moveLeftAct)
        self.addAction(self.moveRightAct)

    def setupMenus(self):
        """
        Creates the menus and adds the actions to them
        """
        fileMenu = self.menuBar().addMenu("&File")
        fileMenu.addAction(self.addRasterAct)
        fileMenu.addMenu(self.vectorMenu)
        fileMenu.addAction(self.removeLayerAct)
        fileMenu.addAction(self.layerAct)
        fileMenu.addAction(self.newWindowAct)
        fileMenu.addAction(self.queryOnlyDisplayedAct)
        fileMenu.addAction(self.tileWindowsAct)
        fileMenu.addAction(self.defaultStretchAct)
        fileMenu.addAction(self.saveCurrentViewAct)
        fileMenu.addAction(self.saveCurrentViewersState)
        fileMenu.addAction(self.loadCurrentViewersState)
        fileMenu.addAction(self.propertiesAct)
        fileMenu.addAction(self.exitAct)
        fileMenu.insertSeparator(self.exitAct)

        editMenu = self.menuBar().addMenu("&Edit")
        editMenu.addAction(self.stretchAct)
        editMenu.addAction(self.preferencesAct)

        viewMenu = self.menuBar().addMenu("&View")
        viewMenu.addAction(self.panAct)
        viewMenu.addAction(self.zoomInAct)
        viewMenu.addAction(self.zoomOutAct)
        viewMenu.addAction(self.zoomNativeAct)
        viewMenu.addAction(self.zoomFullExtAct)
        viewMenu.addAction(self.followExtentAct)
        viewMenu.addAction(self.timeseriesForwardAct)
        viewMenu.addAction(self.timeseriesBackwardAct)

        toolMenu = self.menuBar().addMenu("&Tools")
        toolMenu.addAction(self.queryAct)
        toolMenu.addAction(self.newQueryAct)
        toolMenu.addAction(self.vectorQueryAct)
        toolMenu.addAction(self.newVectorQueryAct)
        toolMenu.addAction(self.profileAct)
        toolMenu.addAction(self.newProfileAct)
        toolMenu.addAction(self.flickerAct)

        helpMenu = self.menuBar().addMenu("&Help")
        helpMenu.addAction(self.aboutAct)


    def setupToolbars(self):
        """
        Creates the toolbars and adds the actions to them
        """
        fileToolbar = self.addToolBar("File")
        fileToolbar.addAction(self.addRasterAct)
        vectorToolButton = QToolButton()
        vectorToolButton.setMenu(self.vectorMenu)
        vectorToolButton.setPopupMode(QToolButton.MenuButtonPopup)
        vectorToolButton.setIcon(QIcon(":/viewer/images/addvector.png"))
        vectorToolButton.setDefaultAction(self.addVectorFileAct)
        fileToolbar.addWidget(vectorToolButton)

        fileToolbar.addAction(self.removeLayerAct)
        fileToolbar.addAction(self.layerAct)
        fileToolbar.addAction(self.newWindowAct)
        fileToolbar.addAction(self.propertiesAct)

        viewToolbar = self.addToolBar("View")
        viewToolbar.addAction(self.panAct)
        viewToolbar.addAction(self.zoomInAct)
        viewToolbar.addAction(self.zoomOutAct)
        viewToolbar.addAction(self.zoomNativeAct)
        viewToolbar.addAction(self.zoomFullExtAct)
        viewToolbar.addAction(self.followExtentAct)

        toolToolbar = self.addToolBar("Tools")
        toolToolbar.addAction(self.queryAct)
        toolToolbar.addAction(self.vectorQueryAct)
        toolToolbar.addAction(self.profileAct)
        toolToolbar.addAction(self.flickerAct)

    def setupStatusBar(self):
        """
        Sets up the status bar
        """
        statusbar = self.statusBar()
        statusbar.setSizeGripEnabled(True)
        self.progressWidget = QProgressBar(statusbar)
        self.progressWidget.setMinimum(0)
        self.progressWidget.setMaximum(100)
        self.progressWidget.setVisible(False)
        statusbar.addPermanentWidget(self.progressWidget)

    def newWindow(self):
        """
        Triggered when user wants a new window. Send signal
        to GeolinkedViewers class (if there is one!)
        """
        self.newWindowSig.emit()

    def tileWindows(self):
        """
        Triggered when user wants to tile windows. Display
        dialog to allow number to be selected, then send signal
        to GeolinkedViewers class (if there is one!)
        """
        from .tiledialog import TileDialog
        dlg = TileDialog(self)
        if dlg.exec_() == QDialog.Accepted:
            xnum, ynum = dlg.getValues()
            self.tileWindowsSig.emit(xnum, ynum)

    def defaultStretch(self):
        """
        Show the default stretch dialog
        """
        from . import stretchdialog
        dlg = stretchdialog.StretchDefaultsDialog(self)
        dlg.exec_()

    def addRaster(self):
        """
        User has asked to open a file. Show file
        dialog and open file
        """
        populateFilters()
        dlg = QFileDialog(self)
        dlg.setNameFilters(GDAL_FILTERS)
        dlg.setFileMode(QFileDialog.ExistingFiles)
        # set last dir
        layer = self.viewwidget.layers.getTopRasterLayer()
        if layer is not None:
            dir = os.path.dirname(layer.filename)
            if dir == '':
                dir = os.getcwd()
        else:
            # or cwd
            dir = os.getcwd()

        dlg.setDirectory(dir)

        if dlg.exec_() == QDialog.Accepted:
            file_list = map(str, dlg.selectedFiles())
            for fname in archivereader.file_list_to_archive_strings(file_list):
                self.addRasterInternal(fname)

    def addVectorFile(self):
        """
        User wants to add a vector layer. OGR seems to have no
        way to determine extensions...
        From a file.
        """
        dlg = QFileDialog(self)
        dlg.setNameFilter("OGR Files (*)")
        dlg.setFileMode(QFileDialog.ExistingFile)
        # set last dir
        layer = self.viewwidget.layers.getTopVectorLayer()
        if layer is not None:
            dir = os.path.dirname(layer.filename)
            if dir == '':
                dir = os.getcwd()
        else:
            # or cwd
            dir = os.getcwd()

        dlg.setDirectory(dir)

        if dlg.exec_() == QDialog.Accepted:
            fname = dlg.selectedFiles()[0]
            self.addVectorInternal(fname)

    def addVectorDir(self):
        """
        Add a vector from a directory (filegdb/covereage)
        """
        # set last dir
        layer = self.viewwidget.layers.getTopVectorLayer()
        if layer is not None:
            olddir = os.path.dirname(layer.filename)
            if olddir == '':
                olddir = os.getcwd()
        else:
            # or cwd
            olddir = os.getcwd()

        dir = QFileDialog.getExistingDirectory(self, "Choose vector directory",
            directory=olddir,
            options=QFileDialog.ShowDirsOnly|QFileDialog.DontResolveSymlinks)
        if dir != "":
            self.addVectorInternal(dir)

    def addVectorDB(self):
        """
        Add a vector from a database - ask user for connection string
        """
        from PyQt5.QtWidgets import QInputDialog
        (con, ok) = QInputDialog.getText(self, MESSAGE_TITLE, 
                                "Enter OGR connection string (without quotes)")
        if ok and con != "":
            self.addVectorInternal(con)

    @staticmethod
    def findDefaultStretchForDataset(gdaldataset):
        """
        Attempts to find the default stretch that matches the
        given gdal dataset. Returns None on failure.
        """
        from . import stretchdialog
        stretch = None
        defaultList = stretchdialog.StretchDefaultsDialog.fromSettings()
        for rule in defaultList:
            if rule.isMatch(gdaldataset):
                stretch = rule.stretch
                break
        return stretch

    def addRasterInternal(self, fname, stretch=None, showError=True):
        """
        Actually to the file opening. If stretch is None
        is is determined using our automatic scheme.
        if showError is True a message box will be displayed with any error
        if false an exception will be raised.
        """
        fname = str(fname) # was QString
        lut = None
        # first open the dataset
        from osgeo import gdal
        try:
            gdal.PushErrorHandler('CPLQuietErrorHandler')
            gdaldataset = gdal.Open(fname)
        except RuntimeError as err:
            if SHOW_TRACEBACK:
                traceback.print_exc()
            if showError:
                msg = "Unable to open %s\n%s" % (fname,err)
                QMessageBox.critical(self, MESSAGE_TITLE, msg)
                return
            else:
                raise

        if stretch is None:
            # first see if it has a stretch saved in the file
            from . import viewerstretch
            stretch = viewerstretch.ViewerStretch.readFromGDAL(gdaldataset)
            if stretch is None:
                # ok was none, read in the default stretches
                stretch = self.findDefaultStretchForDataset(gdaldataset)
            else:
                # if there was a stretch, see if we can read a LUT also
                from . import viewerLUT
                lut = viewerLUT.ViewerLUT.createFromGDAL(gdaldataset, stretch)


            # couldn't find anything. Tell user and
            # open default stretch dialog
            if stretch is None:
                del gdaldataset
                msg = ("File has no stretch saved and none of the default " + 
                "stretches match\nThe default stretch dialog will now open.")
                QMessageBox.warning(self, MESSAGE_TITLE, msg)
                self.defaultStretch()
                return

        # now open it for real
        try:
            self.viewwidget.addRasterLayer(gdaldataset, stretch, lut)
        except viewererrors.ProjectionMismatch as e:
            # as the user if they really want to go ahead
            btn = QMessageBox.question(self, MESSAGE_TITLE, 
                """Projection is different to existing file(s). 
Results may be incorrect. Do you wish to go ahead anyway?""", 
                QMessageBox.Yes | QMessageBox.No, QMessageBox.No)
            if btn == QMessageBox.Yes:
                # try again with the flag
                try:
                    self.viewwidget.addRasterLayer(gdaldataset, stretch, lut,
                            ignoreProjectionMismatch=True)
                except Exception as e:
                    if SHOW_TRACEBACK:
                        traceback.print_exc()
                    if showError:
                        QMessageBox.critical(self, MESSAGE_TITLE, str(e) )
                    else:
                        raise
        except viewererrors.InvalidStretch as e:
            # probably band referred to in stretch no longer exists
            # display error and fall back on default stretch
            QMessageBox.information(self, MESSAGE_TITLE,
                """Saved stretch refers to invalid band(s).
File will now be opened using default stretch""")

            stretch = self.findDefaultStretchForDataset(gdaldataset)
            if stretch is None:
                del gdaldataset
                msg = ("File has no stretch saved and none of the default " + 
                "stretches match\nThe default stretch dialog will now open.")
                QMessageBox.warning(self, MESSAGE_TITLE, msg)
                self.defaultStretch()
                return

            # now call this function again with default stretch
            del gdaldataset
            self.addRasterInternal(fname, stretch=stretch)

        except Exception as e:
            if SHOW_TRACEBACK:
                traceback.print_exc()
            if showError:
                QMessageBox.critical(self, MESSAGE_TITLE, str(e) )
            else:
                raise

        # allow the stretch to be edited
        self.stretchAct.setEnabled(True)

    def addVectorInternal(self, path, layername=None):
        """
        Open OGR dataset and layer and tell widget to add it 
        to the list of layers
        """
        from osgeo import ogr
        isResultSet = False
        origSQL = None
        try:
            ds = ogr.Open(str(path))
            if ds is None:
                msg = 'Unable to open %s' % path
                QMessageBox.critical(self, MESSAGE_TITLE, msg)
                return
                
            if layername is None:
                # ask them
                numLayers = ds.GetLayerCount()
                if numLayers == 0:
                    raise IOError("no valid layers")
                else:
                    from . import vectoropendialog
                    layerNames = []
                    for n in range(ds.GetLayerCount()):
                        name = ds.GetLayer(n).GetName()
                        layerNames.append(name)

                    dlg = vectoropendialog.VectorOpenDialog(self, layerNames)
                    if dlg.exec_() == QDialog.Accepted:
                        if dlg.isNamedLayer():
                            name = dlg.getSelectedLayer()
                            lyr = ds.GetLayerByName(str(name))
                        else:
                            sql = dlg.getSQL()
                            origSQL = str(sql)
                            lyr = ds.ExecuteSQL(origSQL)
                            if lyr is None:
                                raise IOError("Invalid SQL")                                
                            isResultSet = True
                    else:
                        return

            else:
                lyr = ds.GetLayerByName(layername)
                
            self.viewwidget.addVectorLayer(ds, lyr, resultSet=isResultSet,
                                        origSQL=origSQL)

        except Exception as e:
            if SHOW_TRACEBACK:
                traceback.print_exc()
            QMessageBox.critical(self, MESSAGE_TITLE, str(e) )

    def addLayersFromJSONFile(self, fileobj, nlayers):
        """
        Gets the widget to read layer definitions from fileobj and add them
        """
        try:
            self.viewwidget.addLayersFromJSONFile(fileobj, nlayers)
        except Exception as e:
            QMessageBox.critical(self, MESSAGE_TITLE, str(e) )
        self.stretchAct.setEnabled(True)

    def removeLayer(self):
        """
        Remove the top most layer
        """
        self.viewwidget.removeLayer()

    def layerWindowClosed(self, window):
        """
        Reset state so we know we need to open another
        """
        self.layerWindow = None

    def arrangeLayers(self):
        """
        Toggle the LayerWindow
        """
        from . import layerwindow
        if self.layerWindow is None:
            self.layerWindow = layerwindow.LayerWindow(self, self.viewwidget)
            self.addDockWidget(Qt.LeftDockWidgetArea, self.layerWindow)
            # this works to prevent it trying to dock when dragging
            # but double click still works
            self.layerWindow.setAllowedAreas(Qt.NoDockWidgetArea) 
            self.layerWindow.layerWindowClosed.connect(self.layerWindowClosed)
        else:
            # remove
            self.removeDockWidget(self.layerWindow)
            self.layerWindow = None

    def editStretch(self):
        """
        Show the edit stretch dock window
        """
        from . import stretchdialog
        # should it just be visible layers?
        layer = self.viewwidget.layers.getTopRasterLayer()
        if layer is None:
            QMessageBox.critical(self, MESSAGE_TITLE, "No raster layer available" )
        else:
            stretchDock = stretchdialog.StretchDockWidget(self, 
                                self.viewwidget, layer)
            self.addDockWidget(Qt.TopDockWidgetArea, stretchDock)
            # this works to prevent it trying to dock when dragging
            # but double click still works
            stretchDock.setAllowedAreas(Qt.NoDockWidgetArea) 

    def disableTools(self, ignoreTool=None):
        """
        Disable all tool actions apart from ignoreTool
        """
        tools = (self.panAct, self.zoomInAct, self.zoomOutAct, 
                        self.queryAct, self.profileAct, self.vectorQueryAct)
        for tool in tools:
            if tool is not ignoreTool:
                tool.setChecked(False)

    def zoomIn(self, checked):
        """
        Zoom in tool selected.
        Tell view widget to operate in zoom mode.
        """
        if checked:
            # disable any other tools
            self.disableTools(self.zoomInAct)
            self.viewwidget.setActiveTool(viewerwidget.VIEWER_TOOL_ZOOMIN, 
                        id(self))
        elif not self.suppressToolReset:
            self.viewwidget.setActiveTool(viewerwidget.VIEWER_TOOL_NONE, 
                        id(self))

    def zoomOut(self, checked):
        """
        Zoom in tool selected.
        Tell view widget to operate in zoom mode.
        """
        if checked:
            # disable any other tools
            self.disableTools(self.zoomOutAct)
            self.viewwidget.setActiveTool(viewerwidget.VIEWER_TOOL_ZOOMOUT, 
                        id(self))
        elif not self.suppressToolReset:
            self.viewwidget.setActiveTool(viewerwidget.VIEWER_TOOL_NONE, 
                        id(self))

    def pan(self, checked):
        """
        Pan tool selected.
        Tell view widget to operate in pan mode.
        """
        if checked:
            # disable any other tools
            self.disableTools(self.panAct)
            self.viewwidget.setActiveTool(viewerwidget.VIEWER_TOOL_PAN, 
                        id(self))
        elif not self.suppressToolReset:
            self.viewwidget.setActiveTool(viewerwidget.VIEWER_TOOL_NONE, 
                        id(self))

    def moveFixedDist(self, xdist, ydist):
        layer = self.viewwidget.layers.getTopRasterLayer()
        if layer is not None:
            # stop panning and move viewport
            (pixNewX, pixNewY) = layer.coordmgr.display2pixel(xdist, 
                                                            ydist)
            #print 'panning'
            #print layer.coordmgr
            layer.coordmgr.setTopLeftPixel(pixNewX, pixNewY)
            layer.coordmgr.recalcBottomRight()
            #print layer.coordmgr
            # reset
            self.viewwidget.paintPoint.setX(0)
            self.viewwidget.paintPoint.setY(0)
            # redraw
            self.viewwidget.layers.makeLayersConsistent(layer)
            self.viewwidget.layers.updateImages()
            self.viewwidget.viewport().update()
            self.viewwidget.updateScrollBars()
            # geolink
            self.viewwidget.emitGeolinkMoved()

    def moveUp(self):
        self.moveFixedDist(0, -1 * self.viewwidget.size().height())

    def moveDown(self):
        self.moveFixedDist(0, self.viewwidget.size().height())

    def moveLeft(self):
        self.moveFixedDist(-1 * self.viewwidget.size().width(), 0)

    def moveRight(self):
        self.moveFixedDist(self.viewwidget.size().width(), 0)

    def zoomNative(self):
        """
        Tell the widget to zoom to native resolution
        """
        try:
            self.viewwidget.zoomNativeResolution()
        except Exception as e:
            QMessageBox.critical(self, MESSAGE_TITLE, str(e) )

    def zoomFullExtent(self):
        """
        Tell the widget to zoom back to the full extent
        """
        try:
            self.viewwidget.zoomFullExtent()
        except Exception as e:
            QMessageBox.critical(self, MESSAGE_TITLE, str(e) )

    def followExtent(self, state):
        """
        Called when user toggles the follow extent button.
        Tell the widget
        """
        self.viewwidget.setGeolinkFollowExtentAction(state)

    def query(self, checked):
        """
        Query tool selected.
        Tell view widget to operate in query mode.
        """
        if checked:
            # disable any other tools
            self.disableTools(self.queryAct)
            self.viewwidget.setActiveTool(viewerwidget.VIEWER_TOOL_QUERY, 
                    id(self))

            # if there is no query window currently open start one
            if self.queryWindowCount <= 0:
                self.newQueryWindow()
        else:
            self.viewwidget.setActiveTool(viewerwidget.VIEWER_TOOL_NONE, 
                    id(self))

    def queryClosed(self, queryDock):
        """
        Query dock window has been closed. Disconnect from
        locationSelected signal and decrement our count
        """
        self.viewwidget.locationSelected.disconnect(queryDock.locationSelected)
        self.queryWindowCount -= 1

    def newQueryWindow(self):
        """
        Create a new QueryDockWidget and connect signals
        and increment our count of these windows
        """
        queryDock = querywindow.QueryDockWidget(self, self.viewwidget)
        # can't pass Qt.NoDockWidgetArea in here
        self.addDockWidget(Qt.BottomDockWidgetArea, queryDock)
        queryDock.setFloating(True) # detach so it isn't docked by default
        # this works to prevent it trying to dock when dragging
        # but double click still works
        queryDock.setAllowedAreas(Qt.NoDockWidgetArea) 
        
        # start over this window
        thispos = self.pos()
        x = thispos.x() + 100
        y = thispos.y() + 100
        queryDock.move(x, y)

        # connect it to signals emitted by the viewerwidget
        self.viewwidget.locationSelected.connect(queryDock.locationSelected)

        # grab the signal the queryDock sends when it is closed
        queryDock.queryClosed.connect(self.queryClosed)

        # increment our count
        self.queryWindowCount += 1

        # emit the signal back to geolinked viewers so that 
        # any plugins can be informed
        self.newQueryWindowSig.emit(queryDock)

    def vectorQuery(self, checked):
        """
        Vector Query tool selected
        Tell view widget to operate in query mode
        """
        if checked:
            self.disableTools(self.vectorQueryAct)
            self.viewwidget.setActiveTool(viewerwidget.VIEWER_TOOL_VECTORQUERY,
                    id(self))

            # if no window, start one
            if self.vectorQueryWindowCount <= 0:
                self.newVectorQueryWindow()
        else:
            self.viewwidget.setActiveTool(viewerwidget.VIEWER_TOOL_NONE, 
                        id(self))

    def newVectorQueryWindow(self):
        """
        Create a new VectorQueryDockWidget and connect signals
        and increment our count of these windows
        """
        from . import vectorquerywindow
        queryDock = vectorquerywindow.VectorQueryDockWidget(self)
        self.addDockWidget(Qt.BottomDockWidgetArea, queryDock)
        queryDock.setFloating(True) # detach so it isn't docked by default
        # this works to prevent it trying to dock when dragging
        # but double click still works
        queryDock.setAllowedAreas(Qt.NoDockWidgetArea) 

        # start over this window
        thispos = self.pos()
        x = thispos.x() + 100
        y = thispos.y() + 100
        queryDock.move(x, y)

        # connect it to signals emitted by the viewerwidget
        self.viewwidget.vectorLocationSelected.connect(
                            queryDock.vectorLocationSelected)

        # grab the signal the queryDock sends when it is closed
        queryDock.queryClosed.connect(self.vectorQueryClosed)

        # increment our count
        self.vectorQueryWindowCount += 1

        # emit the signal back to geolinked viewers so that 
        # any plugins can be informed
        self.newQueryWindowSig.emit(queryDock)

    def vectorQueryClosed(self, queryDock):
        """
        Query dock window has been closed. Disconnect from
        vectorLocationSelected signal and decrement our count
        """
        self.viewwidget.vectorLocationSelected.disconnect(
                        queryDock.vectorLocationSelected)
        self.vectorQueryWindowCount -= 1

    def profile(self, checked):
        """
        Profile tool selected.
        Tell view widget to operate in polyline mode.
        """
        if checked:
            # disable any other tools
            self.disableTools(self.profileAct)
            self.viewwidget.setActiveTool(viewerwidget.VIEWER_TOOL_POLYLINE, 
                        id(self))

            # if there is no query window currently open start one
            if self.profileWindowCount <= 0:
                self.newProfile()
        else:
            self.viewwidget.setActiveTool(viewerwidget.VIEWER_TOOL_NONE, 
                        id(self))

    def newProfile(self):
        from . import profilewindow
        profileDock = profilewindow.ProfileDockWidget(self, self.viewwidget)
        self.addDockWidget(Qt.TopDockWidgetArea, profileDock)
        profileDock.setFloating(True) # detach so it isn't docked by default
        # this works to prevent it trying to dock when dragging
        # but double click still works
        profileDock.setAllowedAreas(Qt.NoDockWidgetArea) 

        # start over this window
        thispos = self.pos()
        x = thispos.x() + 100
        y = thispos.y() + 100
        profileDock.move(x, y)

        # connect to the signal that provides our new line
        self.viewwidget.polylineCollected.connect(profileDock.newLine)

        # grab the signal the profileDock sends when it is closed
        profileDock.profileClosed.connect(self.profileClosed)

        # increment our count
        self.profileWindowCount += 1

    def profileClosed(self, profileDock):
        """
        Profile dock window has been closed. Disconnect from
        polylineCollected signal and decrement our count
        """
        self.viewwidget.polylineCollected.disconnect(profileDock.newLine)
        self.profileWindowCount -= 1

    def properties(self):
        """
        Show the properties dialog
        """
        from . import propertieswindow
        layer = self.viewwidget.layers.getTopLayer()
        if layer is not None:
            info = layer.getPropertiesInfo()
            dlg = propertieswindow.PropertiesWindow(self, info)
            dlg.setWindowTitle(layer.title)
            dlg.show()

    def flicker(self):
        """
        Tell the widget to flicker
        """
        state = self.viewwidget.flicker()
        if state:
            self.flickerAct.setIcon(self.flickerAct.iconOn)
        else:
            self.flickerAct.setIcon(self.flickerAct.iconOff)

    def saveCurrentView(self):
        """
        Saves the current view as an image file
        """
        # now get a filename
        fname, filter = QFileDialog.getSaveFileName(self, "Image File", 
                        filter="Images (*.png *.xpm *.jpg *.tif)")
        if fname != '':
            self.saveCurrentViewInternal(fname)

    def saveCurrentViewInternal(self, fname):
        """
        Saves the current view as an image file as the file given
        """
        # first grab it out of the widget
        from PyQt5.QtGui import QImage
        img = QImage(self.viewwidget.viewport().size(), QImage.Format_RGB32)
        self.viewwidget.viewport().render(img)

        if not img.save(fname):
            QMessageBox.critical(self, MESSAGE_TITLE, 
                "Unable to save file")
        else:
            # save a world file while we are at it
            # see http://en.wikipedia.org/wiki/World_file
            worldfname = fname + 'w'
            worldfObj = None
            try:
                layer = self.viewwidget.layers.getTopRasterLayer()
                if layer is not None:
                    metresperwinpix = (layer.coordmgr.imgPixPerWinPix * 
                                    layer.coordmgr.geotransform[1])
                    (left, top, right, bottom) = (
                                     layer.coordmgr.getWorldExtent())

                    worldfObj = open(worldfname, 'w')
                    worldfObj.write("%f\n" % metresperwinpix)
                    worldfObj.write("0.0\n0.0\n")
                    worldfObj.write("%f\n" % -metresperwinpix)
                    worldfObj.write("%f\n" % (left + (metresperwinpix / 2.0)))
                    worldfObj.write("%f\n" % (top + (metresperwinpix / 2.0)))
            except IOError as e:
                QMessageBox.critical(self, MESSAGE_TITLE,
                    "Unable to save world file: %s" % s)
            finally:
                if worldfObj is not None:
                    worldfObj.close()

    def about(self):
        """
        Show author and version info
        """
        from . import TUIVIEW_VERSION
        from PyQt5.QtCore import PYQT_VERSION_STR, QT_VERSION_STR
        from osgeo.gdal import __version__ as gdalVersion
        import sys
        from numpy import version as numpyVersion

        msg = """<p align='center'>TuiView<br><br>
By Sam Gillingham, Neil Flood, Pete Bunting, James Shepherd, <br>
Pierre Roudier, Tony Gill, Robin Wilson, Dan Clewley, Dale Peters<br>
and Terry Cain.<br><br>

Development funded by Landcare Research NZ Ltd.<br><br>

Colours from www.colorbrewer.org by Cynthia A. Brewer, Geography,<br>
Pennsylvania State University.<br><br>

Version: %s<br>
Installed in: %s<br>
GDAL Version: %s<br>
PyQt Version: %s<br>
Qt Version: %s<br>
Python Version: %s<br>
Numpy Version: %s<br></p>
"""
        appDir = os.path.dirname(os.path.abspath(sys.argv[0]))
        pyVer = "%d.%d.%d" % (sys.version_info.major, sys.version_info.minor,
                    sys.version_info.micro)
        msg = msg % (TUIVIEW_VERSION, appDir, gdalVersion, PYQT_VERSION_STR, 
                QT_VERSION_STR, 
                pyVer, numpyVersion.version)

        # centre each line - doesn't work very well due to font
        msgLines = msg.split('\n')
        maxLine = max([len(line) for line in msgLines])
        centredMsgs = []
        for line in msgLines:
            leftSpaces = int((maxLine - len(line)) / 2.0)
            centred = (' ' * leftSpaces) + line
            centredMsgs.append(centred)

        QMessageBox.about(self, MESSAGE_TITLE, "\n".join(centredMsgs))


    def closeEvent(self, event):
        """
        Window is being closed. Save the position and size.
        Check that any of the query windows don't have unsaved data
        """
        settings = QSettings()
        settings.beginGroup('ViewerWindow')
        settings.setValue("size", self.size())
        settings.setValue("pos", self.pos())
        settings.endGroup()

        event.accept()

    def dragEnterEvent(self, event):
        """
        Called when user about to drop some data on the window
        accept it if it has urls (which are usually just files)
        """
        mimeData = event.mimeData()
        if mimeData.hasUrls():
            event.acceptProposedAction()

    def dropEvent(self, event):
        """
        Called when the user attempts to drop some data onto the window
        We only respond to files being dropped
        """
        mimeData = event.mimeData()
        if mimeData.hasUrls():
            for url in mimeData.urls():
                fname = url.toLocalFile()
                try:
                    # try raster first
                    self.addRasterInternal(fname, showError=False)
                except Exception as e:
                    # then vector 
                    self.addVectorInternal(fname)

    def setPreferences(self):
        """
        Display the preferences dialog
        """
        from . import viewerpreferences
        viewPref = viewerpreferences.ViewerPreferencesDialog(self)
        if viewPref.exec_() == QDialog.Accepted:

            # extract the mouse wheel setting
            self.mouseWheelZoom = viewPref.settingMouseWheelZoom
            self.viewwidget.setMouseScrollWheelAction(self.mouseWheelZoom)
            # the color
            self.backgroundColor = viewPref.settingBackgroundColor
            self.viewwidget.setBackgroundColor(self.backgroundColor)
            # setting have already been set back to disc

    def queryOnlyDisplayed(self, state):
        "Change the state of the querying behaviour"
        self.viewwidget.setQueryOnlyDisplayed(state)

    def saveViewersState(self):
        """
        Get the geolinked viewers class to save the state as a file
        """
        # set last dir
        layer = self.viewwidget.layers.getTopLayer()
        if layer is not None:
            dir = os.path.dirname(layer.filename)
            if dir == '':
                dir = os.getcwd()
        else:
            # or cwd
            dir = os.getcwd()
        fname, filter= QFileDialog.getSaveFileName(self, 
                    "Select file to save state into",
                    dir, "TuiView State .tuiview (*.tuiview)")

        if fname != "":
            fileobj = open(fname, 'w')
            self.writeViewersState.emit(fileobj)
            fileobj.close()

    def loadViewersState(self):
        """
        Get the geolinked viewers class to open previously saved state file
        """
        # set last dir
        layer = self.viewwidget.layers.getTopLayer()
        if layer is not None:
            dir = os.path.dirname(layer.filename)
            if dir == '':
                dir = os.getcwd()
        else:
            # or cwd
            dir = os.getcwd()
        fname, filter = QFileDialog.getOpenFileName(self, "Select file to restore state from",
                    dir, "TuiView State .tuiview (*.tuiview)")

        if fname != "":
            fileobj = open(fname)
            self.readViewersState.emit(fileobj)
            fileobj.close()





"""
Module that contains ViewerStretch and StretchRule classes
"""
# This file is part of 'TuiView' - a simple Raster viewer
# Copyright (C) 2012  Sam Gillingham
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 2
# of the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.

import copy
import json
from osgeo import gdal
from . import viewererrors

gdal.UseExceptions()

# constants for specifying how to display an image 
VIEWER_MODE_DEFAULT = 0
VIEWER_MODE_COLORTABLE = 1
VIEWER_MODE_GREYSCALE = 2
VIEWER_MODE_RGB = 3
VIEWER_MODE_PSEUDOCOLOR = 4

# how to stretch an image
VIEWER_STRETCHMODE_DEFAULT = 0
VIEWER_STRETCHMODE_NONE = 1 # color table, or pre stretched data
VIEWER_STRETCHMODE_LINEAR = 2
VIEWER_STRETCHMODE_STDDEV = 3
VIEWER_STRETCHMODE_HIST = 4

# for storing a stretch within a file
VIEWER_STRETCH_METADATA_KEY = 'VIEWER_STRETCH'

# default stretchparams
VIEWER_DEFAULT_STDDEV = 2.0
VIEWER_DEFAULT_HISTMIN = 0.025
VIEWER_DEFAULT_HISTMAX = 0.01

VIEWER_DEFAULT_NOCOLOR = (0, 0, 0, 0)

class ViewerStretch(object):
    """
    Class that represents the stretch. 
    Use the methods here to set the type of
    stretch you want and the bands used
    """
    def __init__(self):
        self.mode = VIEWER_MODE_DEFAULT
        self.stretchmode = VIEWER_STRETCHMODE_DEFAULT
        self.stretchparam = None
        self.bands = None
        self.rampName = None  # from colorbrewer2.org
        self.nodata_rgba = VIEWER_DEFAULT_NOCOLOR
        self.background_rgba = VIEWER_DEFAULT_NOCOLOR
        self.nan_rgba = VIEWER_DEFAULT_NOCOLOR
        self.attributeTableSize = None # override with size of attribute table 
                                    # if one exists
                                    # LUT will then be created with this size
        self.readLUTFromText = None # if not None, path to text 
                                    # file to read LUT out of
        self.readLUTFromGDAL = None # if not None, path to GDAL dataset 
                                    # to read LUT out of

    def setBands(self, bands):
        "Set the bands to use. bands should be a tuple of 1-based ints"
        self.bands = bands

    def setColorTable(self):
        "Use the color table in the image"
        self.mode = VIEWER_MODE_COLORTABLE
        self.stretchmode = VIEWER_STRETCHMODE_NONE

    def setGreyScale(self):
        "Display a single band in greyscale"
        self.mode = VIEWER_MODE_GREYSCALE

    def setPseudoColor(self, rampName):
        "Display with given color ramp"
        self.mode = VIEWER_MODE_PSEUDOCOLOR
        self.rampName = rampName

    def setRGB(self):
        "Display 3 bands as RGB"
        self.mode = VIEWER_MODE_RGB

    def setNoStretch(self):
        "Don't do a stretch - data is already stretched"
        self.stretchmode = VIEWER_STRETCHMODE_NONE

    def setLinearStretch(self, minVal=None, maxVal=None):
        """
        Just stretch linearly between min and max values
        if None, range of the data used
        """
        self.stretchmode = VIEWER_STRETCHMODE_LINEAR
        self.stretchparam = (minVal, maxVal)

    def setStdDevStretch(self, stddev=VIEWER_DEFAULT_STDDEV):
        "Do a standard deviation stretch"
        self.stretchmode = VIEWER_STRETCHMODE_STDDEV
        self.stretchparam = (stddev,)

    def setHistStretch(self, minVal=VIEWER_DEFAULT_HISTMIN, 
                            maxVal=VIEWER_DEFAULT_HISTMAX):
        "Do a histogram stretch"
        self.stretchmode = VIEWER_STRETCHMODE_HIST
        self.stretchparam = (minVal, maxVal)

    def setNoDataRGBA(self, rgba):
        "Set the RGBA to display No Data values as"
        self.nodata_rgba = rgba

    def setBackgroundRGBA(self, rgba):
        "Set the RGB to display Background areas as"
        self.background_rgba = rgba

    def setNaNRGBA(self, rgba):
        "Set the RGB to display NaN areas as"
        self.nan_rgba = rgba

    def setAttributeTableSize(self, size):
        """
        set with size of attribute table if one exists
        LUT will then be created with this size
        set to None for default behaviour
        """
        self.attributeTableSize = size

    def setLUTFromText(self, fname):
        "Read in the LUT from specified text file"
        self.readLUTFromText = fname

    def setLUTFromGDAL(self, fname):
        "Read LUT from specified GDAL dataset"
        self.readLUTFromGDAL = fname

    def toString(self):
        """
        Convert to a JSON encoded string
        """
        rep = {'mode' : self.mode, 'stretchmode' : self.stretchmode,
                'stretchparam' : self.stretchparam, 'bands' : self.bands,
                'nodata_rgba' : self.nodata_rgba, 'rampname' : self.rampName,
                'background_rgba' : self.background_rgba, 
                'nan_rgba' : self.nan_rgba}
        return json.dumps(rep)

    @staticmethod
    def fromTextFileWithLUT(fname):
        """
        For reading a .stretch file.
        The stretch is read out of the first line and 
        setLUTFromText called with this file also
        since .stretch files contain both
        """
        fileobj = open(fname)
        s = fileobj.readline()
        fileobj.close()
        stretch = ViewerStretch.fromString(s)
        stretch.setLUTFromText(fname)
        return stretch

    @staticmethod
    def fromGDALFileWithLUT(fname):
        """
        For reading a GDAL file with stretch and LUT
        saved.
        The stretch is read out and
        setLUTFromGDAL called with this file also
        """
        gdaldataset = gdal.Open(fname)
        stretch = ViewerStretch.readFromGDAL(gdaldataset)
        del gdaldataset
        if stretch is not None:
            stretch.setLUTFromGDAL(fname)
        return stretch
    
    @staticmethod
    def fromString(string):
        """
        Create a ViewerStretch instance from a json encoded
        string created by toString()
        """
        rep = json.loads(str(string))

        obj = ViewerStretch()
        obj.mode = rep['mode']
        obj.stretchmode = rep['stretchmode']
        obj.stretchparam = rep['stretchparam']
        obj.bands = rep['bands']
        if 'nodata_rgba' in rep:
            obj.nodata_rgba = rep['nodata_rgba']
        if 'background_rgba' in rep:
            obj.background_rgba = rep['background_rgba']
        if 'nan_rgba' in rep:
            obj.nan_rgba = rep['nan_rgba']
        if 'rampname' in rep:
            obj.rampName = rep['rampname']
        return obj

    def writeToGDAL(self, gdaldataset):
        """
        Write this stretch into the GDAL file
        assumed the dataset opened with GA_Update
        Good idea to reopen any other handles to dataset
        """
        string = self.toString()
        gdaldataset.SetMetadataItem(VIEWER_STRETCH_METADATA_KEY, string)

    @staticmethod
    def deleteFromGDAL(gdaldataset):
        """
        Remove the stretch entry from this dataset
        assumed the dataset opened with GA_Update
        """
        # can't seem to delete an item so set to empty string
        # we test for this explicity below
        gdaldataset.SetMetadataItem(VIEWER_STRETCH_METADATA_KEY, '')

    @staticmethod
    def readFromGDAL(gdaldataset):
        """
        See if there is an entry in the GDAL metadata,
        and return a ViewerStretch instance, otherwise None
        """
        obj = None
        string = gdaldataset.GetMetadataItem(VIEWER_STRETCH_METADATA_KEY)
        if string is not None and string != '':
            obj = ViewerStretch.fromString(string)
        return obj


# Comparison constants to use in StretchRule
VIEWER_COMP_LT = 0 # Less than
VIEWER_COMP_GT = 1 # greater than
VIEWER_COMP_EQ = 2 # equal

class StretchRule(object):
    """
    Class that represents a 'rule' and a stretch
    to be applied when that rule matches.
    The rule contains information about number
    of bands in a dataset and how to compare, plus
    if a dataset has a colour table in a particular band
    """
    def __init__(self, comp, value, ctband, stretch):
        self.comp = comp
        self.value = value
        self.ctband = ctband # or None
        self.stretch = copy.copy(stretch) 
        # can reuse stretch object for something else

    def isMatch(self, gdaldataset):
        """
        Does this rule match the given dataset?
        """
        match = False
        # check band numbers
        if self.comp == VIEWER_COMP_LT:
            match = gdaldataset.RasterCount < self.value
        elif self.comp == VIEWER_COMP_GT:
            match = gdaldataset.RasterCount > self.value
        elif self.comp == VIEWER_COMP_EQ:
            match = gdaldataset.RasterCount == self.value
        else:
            msg = 'invalid value for comparison'
            raise viewererrors.InvalidParameters(msg)

        if (match and self.ctband is not None 
                and self.ctband <= gdaldataset.RasterCount):
            # we match the number of bands
            # but we need to check there is a color 
            # table in the specified band
            gdalband = gdaldataset.GetRasterBand(self.ctband)
            layerType = gdalband.GetMetadataItem('LAYER_TYPE')
            match = False
            # only check if file is thematic anyway
            if layerType is not None and layerType == 'thematic':
                # we really need to check only that the RAT
                # reports that we have the right columns
                hasRed = False
                hasGreen = False
                hasBlue = False
                hasAlpha = False
                rat = gdalband.GetDefaultRAT()
                if rat is not None:
                    ncols = rat.GetColumnCount()
                    for col in range(ncols):
                        usage = rat.GetUsageOfCol(col)
                        if usage == gdal.GFU_Red:
                            hasRed = True
                        elif usage == gdal.GFU_Green:
                            hasGreen = True
                        elif usage == gdal.GFU_Blue:
                            hasBlue = True
                        elif usage == gdal.GFU_Alpha:
                            hasAlpha = True

                match = hasRed and hasGreen and hasBlue and hasAlpha
        
        return match

    def toString(self):
        """
        Convert to a JSON encoded string
        """
        rep = {'comp' : self.comp, 
                'value' : self.value, 'ctband' : self.ctband,
                'stretch' : self.stretch.toString()}
        return json.dumps(rep)

    @staticmethod
    def fromString(string):
        """
        Create a StretchRule instance from a json encoded
        string created by toString()
        """
        rep = json.loads(str(string))

        stretch = ViewerStretch.fromString(rep['stretch'])
        obj = StretchRule(rep['comp'], rep['value'], 
                    rep['ctband'], stretch)
        return obj



"""
Plot widget
"""
# This file is part of 'TuiView' - a simple Raster viewer
# Copyright (C) 2012  Sam Gillingham
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 2
# of the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.

from __future__ import division # ensure we are using Python 3 semantics
import numpy
from PyQt5.QtGui import QPainter, QPainterPath, QPen, QFontMetrics
from PyQt5.QtWidgets import QWidget
from PyQt5.QtCore import Qt, QSize, QSettings

DEFAULT_FONT_SIZE = 6
try:
    DEFAULT_YTICK_FLAGS = Qt.AlignRight | Qt.AlignVCenter | Qt.TextDontClip
    DEFAULT_XTICK_FLAGS = Qt.AlignHCenter | Qt.AlignTop | Qt.TextDontClip
    DEFAULT_LABEL_FLAGS = Qt.AlignLeft | Qt.AlignTop
except TypeError:
    # if building from sphinx these values will be type Mock()
    # which cannot be Or'd. So we fack it.
    DEFAULT_YTICK_FLAGS = 0
    DEFAULT_XTICK_FLAGS = 0
    DEFAULT_LABEL_FLAGS = 0
TICK_SIZE = 2  # pixels

class PlotCurve(object):
    """
    Pass instances of these to PlotLineWidget.addCurve()
    xdata and ydata should be numpy arrays
    If pen not given will be white, 1 pixel wide
    """
    def __init__(self, xdata, ydata, pen=None):
        self.xdata = xdata
        self.ydata = ydata
        if xdata.size == 0 or ydata.size != xdata.size:
            raise ValueError('inavlid data')
        self.pen = pen
        if self.pen is None:
            self.pen = QPen()
            self.pen.setWidth(1)
            self.pen.setColor(Qt.white)

class PlotBars(object):
    """
    Pass an instance of this to PlotBarWidget.setBars()
    data should be a numpy array containing the counts. minVal and maxVal
    describe the range of the data
    """
    def __init__(self, data, minVal, maxVal, pen=None, fillColor=Qt.white):
        self.data = data
        self.minVal = minVal
        self.maxVal = maxVal
        if data.size == 0:
            raise ValueError('inavlid data')
        self.pen = pen
        if self.pen is None:
            self.pen = QPen()
            self.pen.setWidth(1)
            self.pen.setColor(Qt.white)
        self.fillColor = fillColor

class PlotLabel(object):
    """
    Pass instances of these to PlotWidget.addLabel()
    xloc and yloc are in data units
    txt is text to display
    flags are those accepted by QPainter.drawText
    If pen not given will be white, 1 pixel wide
    """
    def __init__(self, xloc, yloc, txt, flags=DEFAULT_LABEL_FLAGS, 
                            pen=None):
        self.xloc = xloc
        self.yloc = yloc
        self.txt = txt
        self.flags = flags
        self.pen = pen
        if self.pen is None:
            self.pen = QPen()
            self.pen.setWidth(1)
            self.pen.setColor(Qt.white)

class PlotTick(object):
    """
    Pass lists of these to PlotWidget.setXTicks and PlotWidget.setYTicks
    if flags is None either DEFAULT_YTICK_FLAGS or DEFAULT_XTICK_FLAGS
    will be used depending on the direction
    if pen is None the default axes pen will be used
    """
    def __init__(self, loc, txt, flags=None, pen=None):
        self.loc = loc
        self.txt = txt
        self.flags = flags
        self.pen = pen

class PlotWidget(QWidget):
    """
    Lightweight plot widget base class
    Don't instantiate directly - use one of the base classes.
    """
    def __init__(self, parent):
        QWidget.__init__(self, parent)
        # always draw background as black
        self.setBackgroundColor(Qt.black)
        self.setAutoFillBackground(True)

        # pen to draw the axes
        self.axesPen = QPen(Qt.gray)
        self.axesPen.setWidth(1)

        # font 
        fontSize = self.getSettingsFontSize()
        self.setFontSize(fontSize)

        # xticks. Lists of PlotTicks
        self.xticks = None
        # yticks
        self.yticks = None

        # text labels list of PlotTexts
        self.labels = []

        # fontmetrics
        self.fontMetrics = QFontMetrics(self.font())

    def haveData(self):
        """
        Returns True if data has been set 
        """
        raise NotImplmentedError('haveData() must be implemented')

    def getSettingsFontSize(self):
        "Get the default font size from settings"
        settings = QSettings()
        settings.beginGroup('Plot')

        fontSize = settings.value('FontSize', DEFAULT_FONT_SIZE, int)

        settings.endGroup()
        return fontSize

    def setFontSize(self, size):
        "Set the font point size"
        font = self.font()
        font.setPointSize(size)
        self.setFont(font)
        self.update()

    def setYRange(self, ymin=None, ymax=None):
        """
        Set the Y range. Pass None for autoscale
        for either or both
        """
        self.yrange = (ymin, ymax)
        self.update()

    def setXRange(self, xmin=None, xmax=None):
        """
        Set the X range. Pass None for autoscale
        for either or both
        """
        self.xrange = (xmin, xmax)
        self.update()

    def setXTicks(self, xticks=None):
        """
        Pass a list of PlotTicks. None to reset.
        """
        self.xticks = xticks
        self.update()

    def setYTicks(self, yticks=None):
        """
        Pass a list of PlotTicks. None to reset.
        """
        self.yticks = yticks
        self.update()

    def setBackgroundColor(self, color):
        "Sets the background color for the widget"
        palette = self.palette()
        palette.setColor(self.backgroundRole(), color)
        self.setPalette(palette)
        self.update()

    def addLabel(self, label):
        "Add a PlotLabel to be drawn"
        self.labels.append(label)
        self.update()

    def removeLabels(self):
        "remove all labels"
        self.labels = []
        self.update()

    def getYDataRange(self):
        """
        Get the range of the Y data to be plotted.
        """
        return self.yrange

    def getXDataRange(self):
        """
        Get the range of the X data to be plotted.
        """
        return self.xrange

    @staticmethod
    def makeIntervals(start, end, nIntervals):
        """
        Make a 'pretty' list of intervals. 
        This was the hardest part.
        also returns number of decimals to display
        This is non-zero if floating point data
        """
        interval = (end - start) / nIntervals

        ndecimals = 0
        test = interval
        if interval >= 1:
            while test > 10:
                test /= 10
                ndecimals += 1
        else:
            while test < 10:
                test *= 10
                ndecimals += 1
            ndecimals = -ndecimals

        newinterval = numpy.ceil(test) * 10**ndecimals
        mininterval = int(start / newinterval) * newinterval

        tmp = mininterval
        intervals = [tmp]
        for n in range(nIntervals):
            tmp += newinterval
            intervals.append(tmp)

        if ndecimals < 0:
            npd = abs(ndecimals)
        else:
            npd = 0

        return intervals, npd

    def drawText(self, paint, xloc, yloc, txt, flags):
        """
        Helper method to draw text in given device coords
        For some reason Qt doesn't come with a method like
        this that moves the text around for given flags.
        """
        # the position we are after is relative to the left bottom of the rect
        txtrect = self.fontMetrics.boundingRect(txt)

        if flags & Qt.AlignRight:
            xloc -= txtrect.width()
        if flags & Qt.AlignHCenter:
            xloc -= txtrect.width() / 2
        if flags & Qt.AlignTop:
            yloc += txtrect.height()
        if flags & Qt.AlignVCenter:
            yloc += txtrect.height() / 2

        if flags & Qt.TextDontClip:
            # remember: y is baseline
            txtrect.setRect(xloc, yloc - txtrect.height(), 
                    txtrect.width(), txtrect.height())

            winrect = self.rect()
            if not winrect.contains(txtrect):
                return

        paint.drawText(xloc, yloc, txt)

    @staticmethod
    def formatInterval(interval, ndp):
        if ndp == 0:
            txt = "%d" % interval
        else:
            txt = "%.*f" % (ndp, interval)
        return txt

    def findLargestIntervalLabel(self, intervals, ndp):
        maxwidth = 0
        for interval in intervals:
            txt = self.formatInterval(interval, ndp)
            textrect = self.fontMetrics.boundingRect(txt)
            txtwidth = textrect.width()
            if txtwidth > maxwidth:
                maxwidth = txtwidth
        return maxwidth

    def drawYTicks(self, paint, minYData, maxYData, yoffset, yscale, height):
        """
        Draw the Y-ticks. Returns the size needed for the text
        """
        paint.setPen(self.axesPen)
        flags = DEFAULT_YTICK_FLAGS

        if self.yticks is None:
            # create our own
            nIntervals = int(height / 20)

            intervals, ndp = self.makeIntervals(minYData, maxYData, nIntervals)

            # find width of largest interval and use that for width param
            txtwidth = self.findLargestIntervalLabel(intervals, ndp)

            for interval in intervals:

                if interval < minYData:
                    continue

                txt = self.formatInterval(interval, ndp)

                yloc = (interval - minYData) * yscale + yoffset

                self.drawText(paint, txtwidth, yloc, txt, flags)
                # draw tick
                paint.drawLine(txtwidth, yloc, txtwidth + TICK_SIZE, yloc)
        else:
            # user supplied

            lastTick = self.xticks[-1]
            textrect = self.fontMetrics.boundingRect(lastTick.txt)
            txtwidth = textrect.width()
            for tick in self.xticks:
                yloc = (tick.loc - minYData) * yscale + yoffset

                if tick.pen is not None:
                    oldPen = paint.pen() # save it
                    paint.setPen(tick.pen)

                flags = DEFAULT_YTICK_FLAGS
                if tick.flags is not None:
                    flags = tick.flags

                self.drawText(paint, txtwidth, yloc, tick.txt, flags)

                if tick.pen is not None:
                    paint.setPen(oldPen) # restore

                # draw tick
                paint.drawLine(txtwidth, yloc, txtwidth + TICK_SIZE, yloc)
    

        return txtwidth + TICK_SIZE

    def drawXTicks(self, paint, minXData, maxXData, xoffset, xscale, width, height):
        """
        Draw the X-ticks
        """
        paint.setPen(self.axesPen)
        flags = DEFAULT_XTICK_FLAGS

        if self.xticks is None:
            # we have to create our own

            # do a guess and refine it until it fits
            txtwidth = 20
            finished = False
            nIntervals = int(width / txtwidth)
            while not finished:

                intervals, ndp = self.makeIntervals(minXData, maxXData, nIntervals)
                txtwidth = self.findLargestIntervalLabel(intervals, ndp)

                nIntervals = int(width / txtwidth)
                # had to make it nIntervals+1 otherwise ended up
                # with infinite loop for long labels for some reason...
                finished = len(intervals) <= nIntervals+1
                
            for interval in intervals:
                if interval < minXData:
                    continue

                txt = self.formatInterval(interval, ndp)

                xloc = (interval - minXData) * xscale + xoffset

                self.drawText(paint, xloc, height, txt, flags)

                # draw tick
                paint.drawLine(xloc, height, xloc, 
                                height + TICK_SIZE)
        else:
            # user supplied ticks
            for tick in self.xticks:
                xloc = (tick.loc - minXData) * xscale + xoffset

                if tick.pen is not None:
                    oldPen = paint.pen() # save it
                    paint.setPen(tick.pen)

                flags = DEFAULT_XTICK_FLAGS
                if tick.flags is not None:
                    flags = tick.flags

                self.drawText(paint, xloc, height, tick.txt, flags)

                if tick.pen is not None:
                    paint.setPen(oldPen) # restore
                # draw tick
                paint.drawLine(xloc, height, xloc, 
                                height + TICK_SIZE)

    def drawLabels(self, paint, minXData, minYData, xoffset, xscale, yoffset, yscale):
        """
        Draw the user supplied labels onto the plot
        """
        for label in self.labels:
            xloc = (label.xloc - minXData) * xscale + xoffset
            yloc = (label.yloc - minYData) * yscale + yoffset
            paint.setPen(label.pen)
            self.drawText(paint, xloc, yloc, label.txt, label.flags)

    def paintEvent(self, event):
        """
        This is the main part - calculation and drawing happen here.
        In theory the calculation should happen separately on resize etc 
        and paint should be simpler, but can't be bothered right now
        Delegates to paintData() in sublass to do actuall work
        """
        paint = QPainter(self)

        # allow enough size under the x-axes for text and tick
        axes_ysize = self.fontMetrics.height() + TICK_SIZE

        size = self.size()
        plotheight = size.height() - axes_ysize

        yoffset = size.height() - axes_ysize
        axes_xsize = axes_ysize # in case there no data, we still have axes drawn ok

        # do we have data?
        if self.haveData():
            minYData, maxYData = self.getYDataRange()
            minXData, maxXData = self.getXDataRange()
            xrange = (maxXData - minXData)
            yrange = (maxYData - minYData)

            # check we can draw lines 
            # - might still be a problem if range set by user
            if xrange > 0 and yrange > 0:

                # NB: Qt works from top left, plots from bottom left
                yscale = -plotheight / yrange

                # axes labels
                axes_xsize = self.drawYTicks(paint, minYData, maxYData, yoffset, 
                                    yscale, plotheight)

                # now we now the width of the axes_xsize calc the other parts
                plotwidth = size.width() - axes_xsize
                xoffset = axes_xsize
                xscale = plotwidth / xrange

                self.drawXTicks(paint, minXData, maxXData, xoffset, 
                                    xscale, plotwidth, plotheight)

                # delegate to sublass
                self.paintData(paint, minXData, minYData, xoffset, xscale, yoffset, yscale)

                # labels
                self.drawLabels(paint, minXData, minYData, xoffset, xscale, yoffset, yscale)

        # axes
        paint.setPen(self.axesPen)
        paint.drawLine(axes_xsize, 0, axes_xsize, size.height() - axes_ysize)
        paint.drawLine(axes_xsize, size.height() - axes_ysize, size.width(), 
                        size.height() - axes_ysize)

        paint.end()

    def paintData(self, paint):
        """
        To be implemented in sublass
        """
        return NotImplementedError('must implement paintData')

    def sizeHint(self):
        """
        This has to be implemented otherwise plot is very small!
        """
        return QSize(400, 400)

class PlotLineWidget(PlotWidget):
    """
    Widget for making line plots
    """
    def __init__(self, parent):
        PlotWidget.__init__(self, parent)

        # default ranges - y autoscale
        self.setYRange()
        # x left =0, right autoscale
        self.setXRange(xmin=0)

        # list of PlotCurves to draw
        self.curves = []

    def haveData(self):
        """
        Returns True if data has been set 
        """
        return len(self.curves) != 0

    def addCurve(self, curve):
        """
        Add an instance of PlotCurve to the list of curves
        to be plotted
        """
        self.curves.append(curve)
        self.update()

    def removeCurves(self):
        "Remove all the curves"
        self.curves = []
        self.update()

    def getYDataRange(self):
        """
        Get the range of the Y data to be plotted.
        If value(s) have been set with SetYRange these
        are returned
        """
        (minYData, maxYData) = PlotWidget.getYDataRange(self)
        if minYData is None:
            minYData = self.curves[0].ydata.min()
            for curve in self.curves[1:]:
                my = curve.ydata.min()
                if my < minYData:
                    minYData = my
        if maxYData is None:
            maxYData = self.curves[0].ydata.max()
            for curve in self.curves[1:]:
                my = curve.ydata.max()
                if my > maxYData:
                    maxYData = my

        if (maxYData - minYData) == 0:
            # make range +/- 20%
            minYData = maxYData - (maxYData * 0.2)
            maxYData = maxYData + (maxYData * 0.2)

        return minYData, maxYData

    def getXDataRange(self):
        """
        Get the range of the X data to be plotted.
        If value(s) have been set with SetXRange these
        are returned
        """
        (minXData, maxXData) = PlotWidget.getXDataRange(self)
        if minXData is None:
            minXData = self.curves[0].xdata.min()
            for curve in self.curves[1:]:
                mx = curve.xdata.min()
                if mx < minXData:
                    minXData = mx
        if maxXData is None:
            maxXData = self.curves[0].xdata.max()
            for curve in self.curves[1:]:
                mx = curve.xdata.max()
                if mx > maxXData:
                    maxXData = mx

        if (maxXData - minXData) == 0:
            # make range +/- 20%
            minXData = maxXData - (maxXData * 0.2)
            maxXData = maxXData + (maxXData * 0.2)

        return minXData, maxXData

    def paintData(self, paint, minXData, minYData, xoffset, xscale, yoffset, yscale):
        """
        Paints the curves - called from paintEvent in the baseclass
        """
        # each curve
        for curve in self.curves:
            paint.setPen(curve.pen)

            xpoints = (curve.xdata - minXData) * xscale + xoffset
            ypoints = (curve.ydata - minYData) * yscale + yoffset
            xpoints = xpoints.astype(int)
            ypoints = ypoints.astype(int)

            # doesn't seem to be a faster array way, but this
            # seems plenty fast enough
            path = QPainterPath()
            path.moveTo(xpoints[0], ypoints[0])
            for x, y in zip(xpoints[1:], ypoints[1:]):
                path.lineTo(x, y)
            paint.drawPath(path)


class PlotBarWidget(PlotWidget):
    """
    Widget for making bar plots
    """
    def __init__(self, parent):
        PlotWidget.__init__(self, parent)

        # default ranges - y min = 0
        self.setYRange(ymin=0)
        # x  autoscale
        self.setXRange()

        # out one PlotBars object
        # maybe we should support more than one?
        self.bars = None

    def haveData(self):
        """
        Returns True if data has been set 
        """
        return self.bars is not None

    def setBars(self, bars):
        "Set the bars object to use"
        self.bars = bars
        self.update()

    def getYDataRange(self):
        """
        Get the range of the Y data to be plotted.
        If value(s) have been set with SetYRange these
        are returned
        """
        (minYData, maxYData) = PlotWidget.getYDataRange(self)
        if minYData is None:
            minYData = self.bars.data.min()
        if maxYData is None:
            maxYData = self.bars.data.max()

        if (maxYData - minYData) == 0:
            # make range +/- 20%
            minYData = maxYData - (maxYData * 0.2)
            maxYData = maxYData + (maxYData * 0.2)

        return minYData, maxYData

    def getXDataRange(self):
        """
        Get the range of the X data to be plotted.
        If value(s) have been set with SetXRange these
        are returned
        """
        (minXData, maxXData) = PlotWidget.getXDataRange(self)
        if minXData is None:
            minXData = self.bars.minVal
        if maxXData is None:
            maxXData = self.bars.maxVal

        if (maxXData - minXData) == 0:
            # make range +/- 20%
            minXData = maxXData - (maxXData * 0.2)
            maxXData = maxXData + (maxXData * 0.2)

        return minXData, maxXData

    def paintData(self, paint, minXData, minYData, xoffset, xscale, yoffset, yscale):
        """
        Paints the bars - called from paintEvent in the baseclass
        """
        paint.setPen(self.bars.pen)
        heights = (self.bars.data - minYData) * yscale + yoffset
        tlxs = numpy.linspace(self.bars.minVal, self.bars.maxVal, self.bars.data.size)
        tlxs = (tlxs - minXData) * xscale + xoffset
        tlxs += 1 # otherwise left most bar gets clobbered by axis
        width = int(tlxs[1] - tlxs[0])
        if width < 1:
            width = 1

        for height, tlx in zip(heights, tlxs):
            paint.fillRect(tlx, height, width, yoffset - height, self.bars.fillColor)
